"""
   General definitions like index grids,...

   COPYRIGHT

     Copyright (c) 2011, Ralf Hambach. All rights reserved.
     Use of this source code is governed by a BSD-style license that 
     can be found in the LICENSE file.
"""
__version__ = "$Revision: 366 $"
__author__ = "Ralf Hambach"
__date__ = "$Date: 2011-11-10 12:34:46 +0100 (Thu, 10 Nov 2011) $"



import numpy as np
from numpy.linalg  import norm
import re;
import string;

class VGrid:
  "Array of arbitrarily distributed Points in dim-dimensions"
  def __init__(self,aPoints,label=('',),unit=('',)):
    """
    vgrid ... Array of dim-dimensional Points with shape (D,:)
    label ... ('labelX','labelY',...,'labelZ') # label for each dim.
    unit  ... ('unitX','unitY',...,'unitZ')    # unit for each dim
    """
    self.dim     = aPoints.shape[0];
    self.shape   = aPoints.shape[1:];
    self.size    = np.prod(self.shape);
    self.vgrid   = np.array(aPoints,dtype=float);
    self.min     = np.amin(self.vgrid.reshape(self.dim,-1),axis=1);
    self.max     = np.amax(self.vgrid.reshape(self.dim,-1),axis=1);
    self.label   = (tuple(label)*self.dim)[0:self.dim];
    self.unit    = (tuple(unit) *self.dim)[0:self.dim];
    
  def extent(self):
    return (self.max-self.min);
  
  def asarray(self):
    """return list of points with shape (self.size, self.dim)"""
    return self.vgrid.reshape(self.dim,self.size).T;

  def flatten(self):
    """return list of points with shape (self.size, self.dim)"""
    return self.vgrid.flatten();

  def index(self,point):
    """index([x,y,...,z]) returns the index-array [i,j,...,k] for point [x,y,...,z]
       NOTE: for 1D-arrays, use index([x])
    """
    index=np.empty((self.dim,),int);
    extent=self.extent();
    for d in range(self.dim):
      if (extent[d]==0):
        index[d]=0;
      else:
        pos=(point[d]-self.min[d])/extent[d];
        index[d]=np.round((self.shape[d]-1)*pos);
    return index;

  def point(self,index):
    """point([i,j,...,k]) returns the point [x,y,...,z] for index-array [i,j,...,k]
       NOTE: for 1D-arrays, use point([i])                                         """
    point=[];
    for component in self.vgrid:
      point.append(component[tuple(index)]);
    return np.array(point,float);


class RegularVGrid(VGrid):
  "List of N regularly distributed Grid Points (vectors of dimension dim)"
  def __init__(self,aN,aLimits,label=('',),unit=('',)):
    """ aN       = [x,y,...,z]         # number of grid points in each dimension
        aLimits  = [[xmin, xmax],      # limits for each dimension
                    [ymin, ymax],
                        ...
                    [zmin, zmax]]
                    
        label = ('labelX','labelY',...,'labelZ') # label for each dim.

        unit  = ('unitX','unitY',...,'unitZ')    # unit for each dim
    """
    self.dim   = len(aN);
    self.shape = tuple(aN);
    self.size  = np.prod(self.shape);   # total number of points
    self.min   = np.array(aLimits,float)[:,0];
    self.max   = np.array(aLimits,float)[:,1];
    self.label = (tuple(label)*self.dim)[0:self.dim];
    self.unit  = (tuple(unit) *self.dim)[0:self.dim];

    # create vector mesh grid of shape (self.dim, self.shape)
    self.vgrid = np.indices(self.shape, float); # contains [0,...,shape-1]
    for d in range(self.dim):          
      if (self.shape[d]==1):
        # set min[d] for all points
        self.vgrid[d]=self.min[d];   
      else:
        # scale to interval [min,max]
        self.vgrid[d]=( self.vgrid[d]/float(self.shape[d]-1) \
                         *(self.max[d]-self.min[d])+self.min[d] );


  def get_line(self,index,axis):
    """return list of indices of points along a line containing the
        point specified by <index> parallel to axis <axis> """
    line=[]; new_index=np.array(index);
    for i in range(self.shape[axis]):
      new_index[axis]=i;
      line.append(new_index.copy());  # deep-copy required
      
    return np.array(line,int);
    
  def get_plane(self,index,axis):
    """return list of indices of points in a plane containing the 
       point specified by <index> perpendicular to axis <axis> """
    shape = list(self.shape); shape.pop(axis);
    plane = np.indices(shape).reshape(self.dim-1,np.prod(shape)).T;
    return np.insert(plane,axis,index[axis],axis=1);

class Atom:
  "Atom(type,pos,color=1), e.g. Atom('C',[1,2,3],2);"
  def __init__(self,type,pos,color=1):
    self.type=type;
    self.pos =np.array(pos,dtype=float);
    self.color=float(color);

  def __str__(self):
    return "type %s: "%str(self.type)+ \
           "pos(%6.3f, %6.3f, %6.3f) "%tuple(self.pos)+\
           "color:%f"%self.color;

  def copy(self):
    "Copy-Constructor"
    return Atom(self.type,self.pos,self.color);

  def print_xyz(self):
    "return string as needed for xyz-files"
    return self.type + " %8.5g %8.5g %8.5g " % tuple(self.pos); 



def epsilon_RPA(chi0,qred_rc,gvec,gprimd):
  " calculate  eps=1-v*chi0,   v=4pi/q^2 "
  qabs = np.asarray([ norm( np.dot(qred_rc+gvec[ig], gprimd) ) for ig in range(len(gvec))]);
  coul = 4*np.pi/qabs**2;
  eps  = - np.multiply(coul, np.asarray(chi0,dtype=complex).T).T \
         + np.identity(len(gvec)); # multiply(): c(i,j)= a(j)*b(i,j)

  return eps
#end epsilon_RPA

def get_shells(cube_G_grid, gvec, gprimd, npwmat=None):
  " re-index list of G-vectors appropriately + select smallest npwmat ones "
  cube_gvec=cube_G_grid.asarray();  
  cube_glen=np.sqrt(np.sum(np.dot(cube_gvec, gprimd)**2,axis=1));
  ig_sorted =np.argsort(cube_glen);
  cube_glen=cube_glen[ig_sorted];

  # determine shells (num. of G-vectors) for distant system
  cube_shells=np.where(cube_glen[1:]-cube_glen[:-1]>1e-6)[0]+1;

  # determine shell corresponding to largest shell of the input chi0
  igmax,     = np.where(np.all(cube_gvec==gvec[-1],axis=1)); 
  cube_maxsh=cube_shells[np.where(cube_shells>igmax)][0];
  if npwmat is None: npwmat = cube_maxsh; 
  
  # test for closed shell (distant system, differ from DP-input shells!)
  if ( npwmat<cube_G_grid.size and 
    cube_glen[npwmat-1]-cube_glen[npwmat]>-1e-6 ):
    raise ValueError("ERROR! shell not closed at npwmat=%d"%npwmat \
     +"\n list of closed shells  :\n" + np.array_str(cube_shells) \
     +"\n proposed shell (for distant sheets) : %d\n"%cube_maxsh);
  if ( npwmat>cube_maxsh):
    raise ValueError("ERROR! including open shells beyond largest "\
     +"G-vector included in DP input file. Use npwmat <= %d instead." % (cube_maxsh));

  cube_ig = ig_sorted[0:npwmat];
  cube_gvec=cube_gvec[cube_ig];

  return (cube_ig,cube_gvec);

def read_lda(infile,ngvec,verbose=False):

  # read rho
  rho_dat=np.fromfile(infile,dtype=float,count=2*ngvec,sep=" ");
  rho=rho_dat[0::2]+1.j*rho_dat[1::2];

  # check for comment line
  line=infile.readline();
  if line.find("# fxc(G,G'') follow")<0: raise IOError('Unexpected file format!');

  # read fxc
  fxc=np.empty((ngvec,ngvec),dtype=complex);
  for ig in range(ngvec):
    for igp in range(ngvec):
      fxc[ig,igp]=complex(*map(float,infile.readline().split(" ")));

  return (rho,fxc)
# end read_lda


########################################################3
# the following code was moved to dp_chi0.py
  
# read num values ignoring line breaks:
def read_wrapped_line(file,num):
  line = read_hline(file).split();
  while (len(line)<num):
    line.extend(read_hline(file).split());
  if (len(line)>num):
    print "FORMAT ERROR! read_wrapped_line() got more values than expected.";
    print line;
  return line;

# read lines ignoring comments
def read_hline(file):
  while True:
    line = file.readline();
    if not re.match("^\s*#",line):
      return line;

# print header
def print_header(ngvec,nomega,q0,q_red,Emin,Emax,dE,rprimd,gprimd,gvec):
  print " Crystal structure ";
  print "   lattice vectors [a.u.]:            "; print(rprimd);
  print "   reciprocal lattice vectors [a.u.]: "; print(gprimd);
  print " ";
  print " DP-parameters ";
  print "   energy: %5.2f - %5.2f eV, dE=%5.2f eV "%(Emin, Emax, dE);
  print "   q [rc]: ",q0;
  print "   q [cc]: ",np.dot(q0,gprimd);
  print "   ngvec : ",ngvec
  print " ";
# end(print_header)

def read_header(file,verbose=False):
  # check for EOF
  line = file.readline() 
  if not line : return None;
  
  # calculation parameter
  [ngvec,nomega] = map(int,  line.split())
  [Emin,Emax,dE] = map(float,file.readline().split())
  
  # polarization vector (TODO: print q_red additionally i.o. to be sure)
  q0      =  np.array(map(float,file.readline().split()))
  q_red   =  q0-np.ceil(np.abs(q0)-[0.5,0.5,0.5])*np.sign(q0);  # wrap to [-0.5,0.5] as in DP

  # crystal structure
  #                    rprimd = [[a1],    gprimd = [[b1],
  #                    in [au]   [a2],    in [au]   [b2],      
  #                              [a3]]              [b3]]   
  rprimd=np.empty((3,3),float)
  for i in range(3):
    rprimd[i] =  np.array(map(float,file.readline().split()))
  gprimd      =  np.array(2*np.pi*np.matrix(rprimd).I.T)
    # a_i.b_j=2pi*delta_ij

  # reciprocal lattice vectors
  gvec = np.array(map(float,read_wrapped_line(file,3*ngvec))).reshape(ngvec,3)

  # define Energies
  E = RegularVGrid([nomega],[[Emin,Emax]],('Energy',),('eV',));
  if verbose:
    print "\n--- read_header ---------------------------------- ";
    print_header(ngvec,nomega,q0,q_red,Emin,Emax,dE,rprimd,gprimd,gvec);
  else:
    print "  q [rc]: ",q0;

  
  return(ngvec,E,q0,q_red,rprimd,gprimd,gvec)
# end(read_header)



# read the header of file SRDENchi
def read_header_new(file,verbose=False):
  # check for EOF
  line = file.readline() 
  if not line : return None;

  # determine version:
  if re.match("^\s*\# -- Header",line): 
    version=int(re.findall("^\s*\# -- Header \(version\s+(\d+)\)",line)[0]);
  
  # calculation parameter
  [ngvec,nomega] = map(int,  read_hline(file).split())
  [Emin,Emax,dE] = map(float,read_hline(file).split())
  
  # polarization vector (TODO: print q_red additionally i.o. to be sure)
  q0      =  np.array(map(float,read_hline(file).split()))
  q_red   =  q0-np.ceil(np.abs(q0)-[0.5,0.5,0.5])*np.sign(q0);  # wrap to [-0.5,0.5] as in DP
  #q_red   =  q0-np.floor(q0+[0.5,0.5,0.5]);  # wrap to [-0.5,0.5) as in DP

  # crystal structure
  #                    rprimd = [[a1],    gprimd = [[b1],
  #                    in [au]   [a2],    in [au]   [b2],      
  #                              [a3]]              [b3]]   
  rprimd=np.empty((3,3),float)
  for i in range(3):
    rprimd[i] =  np.array(map(float,read_hline(file).split()))
  gprimd      =  np.array(2*np.pi*np.matrix(rprimd).I.T)
    # a_i.b_j=2pi*delta_ij

  if (version>0): # atomic positions and types
    natoms = int(read_hline(file));
    T      = []; 
    for i in range(natoms):
      line=read_hline(file).split();
      T.append(Atom(line[0],line[1:4],line[0]));
      
  # reciprocal lattice vectors
  gvec = np.array(map(float,read_wrapped_line(file,3*ngvec))).reshape(ngvec,3)

  # read symmetries (!!!CHECK!!!)
  nsym = int(read_hline(file));
  sym = np.fromfile(file,dtype=int,count=3*3*nsym,sep=" ").reshape(nsym,3,3);
    
  # read polarization
  nqpol = int(read_hline(file));

  # skip rest of header
  if not re.match("^\s*\# --------------",file.readline()): print "ERROR: expected end of header";
  file.readline();

  # define Energies
  E = RegularVGrid([nomega],[[Emin,Emax]],('Energy',),('eV',));
  if verbose:
    print "\n--- read_header (version ",version,")-------------------------- ";
    print_header(ngvec,nomega,q0,q_red,Emin,Emax,dE,rprimd,gprimd,gvec);
  #else:
  #  print "   q [rc]: ",q0;
    
  return(ngvec,E,q0,q_red,rprimd,gprimd,gvec,T,sym,nqpol)
# end(read_header_new)

# check input file:
#  - QMAX
#  - k-points 



def check_input(filename,nskip):
  import matplotlib.pylab as plt; 

  qX=[]; qY=[]; q=[];

  # read parameter from file
  infile = open(filename,'rb')
  while True:
    # read header
    res=read_header_new(infile,verbose=True)
    if res is None:  break
    else: (ngvec,E,q0,q_red,rprimd,gprimd,gvec,T,sym,nqpol)=res;

    q += [ q0 ];

    # skip chi0
    for ig in range(nskip):
      line=read_wrapped_line(infile,2*E.size)
  infile.close();

  print "\n--- check_input ---------------------------------- ";

  # check for QMAX
  qmax=np.max( map(lambda x: norm(x), np.dot(q,gprimd)) );
  Gmax=norm(np.dot(gvec[ngvec-1],gprimd));
  print "  inputfile '%s' contains %d G-vectors, with max|G|=%g, max|q|=%g."\
    %(filename,ngvec,Gmax,qmax);
  print "  ---> QMAX [a.u.] = %5.2f (1/bohr)"%(Gmax-qmax);

  # apply symmetries
  for sym in [-1,1]:
    qp=np.dot(q,gprimd)*sym;
      
    for ig in [-1,0,1]:
      for igp in [-1,0,1]:
        qX+= map(lambda q: (q + ig*gprimd[0])[0], qp);
        qY+= map(lambda q: (q +igp*gprimd[1])[1], qp);

  # count doubled k-points
  qC=np.zeros(np.size(qX));
  for n in range(np.size(qX)):
    x=np.round(qX[n]*1000);
    y=np.round(qY[n]*1000);
 
    points=( np.array([qX,qY]).T * 1000 ).round();    
    for p in points:
      if(p[0]==x and p[1]==y):
        qC[n]+=1;

  import matplotlib.pylab as plt;
  plt.scatter(qX,qY,40,qC);

  return (qX,qY,qC);
# end(check_input)



def read_chi0(infile,ngvec,nqpol,E,verbose=False):

  # read chi0
  chi0=np.empty((ngvec,ngvec,E.size),dtype=complex);
  for ig in range(ngvec):
    for igp in range(ngvec):
      line=np.fromfile(infile,dtype=float,count=2*E.size,sep=" ");
      for io in range(E.size):
        chi0[ig,igp,io]=complex(line[2*io],line[2*io+1]);

  # read lwing, uwing if present
  line=infile.readline();
  if line.find("# lwing, and uwing")>=0:
    lwing=np.empty((nqpol,ngvec,E.size),dtype=complex);
    uwing=np.empty((nqpol,ngvec,E.size),dtype=complex);
    for iqpol in range(nqpol):
      iq=int(re.findall("^\s*\# lwing, and uwing for polarization \s*(\d+)",line)[0]);
      if iq<>iqpol+1: raise RuntimeError("WARNING: unexpected file format!");

      for ig in range(ngvec):
        line1=np.fromfile(infile,dtype=float,count=2*E.size,sep=" ");
        line2=np.fromfile(infile,dtype=float,count=2*E.size,sep=" ");
        for io in range(E.size):
          lwing[iqpol,ig,io]=complex(line1[2*io],line1[2*io+1]);
          uwing[iqpol,ig,io]=complex(line2[2*io],line2[2*io+1]);
      line=infile.readline();
      
    # BUG-correction: for lwing,uwing
    print " Correct DP-Bug: change sign of lwing, uwing "
    lwing[:,1:ngvec,:]*= -1; uwing[:,1:ngvec,:]*= -1;

    # copy wings to chi0 if nqpol=1 (k.p with one selected q)
    if nqpol==1:
      chi0[:,0,:] = lwing[0,:,:];
      chi0[0,:,:] = uwing[0,:,:];
    

  if verbose: print "-------------------------------------------------------"
  if nqpol>1: return (chi0,lwing,uwing)
  else:       return (chi0,)
# end read_chi0




def chi0_to_hypercube(chi0,gvec,Gmax=(None,None,None),verbose=True):
  """
  Re-index the chi0 matrix from DP by the three components of 
  the G-vectors, instead of the linear index used in DP:

   chi0(ig,igp) -->  chi0_cube(iG1,iG2,iG3; iG1',iG2',iG3')

  The origin is in the center of the cube. By default, the size of the
  cube is chosen such that it contains all G-vectors in gvec (they
  form a sphere in cartesian coord.). If one of the Gmax components is
  given, the cube contains G-vectors with components -Gmax[i],..0,..Gmax[i]

  INPUT
   chi0 ... 2D-array chi0(ngvec,ngvec) 
   gvec ... G-vectors gvec(3,ngvec) in r.c.
   Gmax ... (opt)  [Gmax1,Gmax2,Gmax3] specifies the 
            size of the cube for each direction in r.c.

  OUTPUT
   chi0_cube ... 6D-array chi0_cube(nG1,nG2,nG3; nG1,nG2,nG3)
   G_grid    ... corresponding G-vectors
  """
  chi0=np.asarray(chi0); gvec=np.asarray(gvec); Gmax=np.asarray(Gmax);

  # define cube dimensions and reduce chi0 accordingly
  ig = np.ones(chi0.shape[0],dtype=bool);
  for i in range(3):
    if Gmax[i] is None:
      Gmax[i] = np.abs(gvec[:,i]).max().round();
    else:
      if verbose: print('WARNING in chi0_to_hypercube(): restrict G-vectors for chi0 to |G[%d]| <= %d' % (i,Gmax[i]));
      ig &= np.abs(gvec[:,i]) <= Gmax[i];  # False for each G-vector outside the box
  if verbose: print('reshaping chi0 to hypercube with Gmax = ( %3d, %3d, %3d)'%tuple(Gmax));
  
  chi0 = chi0[np.ix_(ig,ig)];
  gvec = gvec[ig];
  ngvec= gvec.shape[0];

  # rectangular, symmetric cube around G-vectors
  Gmax  = np.asarray(Gmax,dtype=int);
  extent= np.array([-Gmax,Gmax]);
  num   = 2*Gmax+1;
  G_grid =RegularVGrid(num,extent.T,('G1','G2','G3'),('G [r.c.]',));
  G_index=np.array([ G_grid.index(gvec[ig]) for ig in range(ngvec) ]);

  # re-index chi0_AO(ig,ig') as chi0_AO(G1,G2,G3,G1',G2',G3');
  chi0_cube=np.zeros((G_grid.shape+G_grid.shape),dtype=complex);
  for ig in range(ngvec):
    for igp in range(ngvec):
      # build index for chi0: might be faster
      index=G_index[ig].tolist()+G_index[igp].tolist();
      chi0_cube[tuple(index)] = chi0[ig,igp]; # copy ?

  return(chi0_cube,G_grid)
#end chi0_to_hypercube



# --- self-test -------------------------------------------------------------

if __name__ == '__main__':

  # class RegularVGrid(VGrid)
  Gmax  = np.array([2,2,5]);
  G0    = np.array([-1,1,1]);
  axis  = 2;
  extent= np.array([-Gmax,Gmax]);
  num   = 2*Gmax+1;
  G_grid=RegularVGrid(num,extent.T,('G1','G2','G3'),('G [r.c.]',));

  # -- get_plane()
  iG_plane = G_grid.get_plane(G_grid.index(G0),axis);
  gvec_plane = np.asarray([G_grid.point(iG) for iG in iG_plane]);

  gvec = G_grid.asarray();
  gvec_plane_comp = gvec[np.where(np.abs(gvec - G0)[:,axis] < 0.1)];

  if not np.all(gvec_plane == gvec_plane_comp):
    raise Exception("Error! Test for RegularVGrid.get_plane() failed.")


  print("All tests passed successfully.")
