#!/usr/bin/python
"""                                                             
     Plot XY-spectra using xmgrace

   VERSION

     $Id: dp2agr.py 343 2011-05-18 23:06:46Z hambach $

   USAGE

     The plot is created in four steps:

     1. One creates a list of spectra (MDF-objects) by
        (a) reading data form all files that match a given pattern
        (b) applying filters to this list
        (c) sorting the list

     2. For each of the MDF-objects in the list a XY-graph is added 
        to the figure. The following functions should be adapted 
        to determine the style of the graph for a given MDF-object
        (e.g. depending on the calculation parameters in the 
        dictionary mdf.param, see MDF interface below):

         get_x(mdf)        ... returns 1D-array of x-data
         get_y(mdf)        ... returns 1D-array of y-data
         get_legend(mdf)   ... returns string for the legend (may be "")
         get_linestyle(mdf)... returns number for the line style
         get_linecolor(mdf)...                             color

      3. The figure properties (Title, Axes,...) are set.

      4. (opt) The figure might be saved to an *.agr, *.ps, ... file

   MDF interface:
   
    member functions: get_E(), get_Reps(), get_Ieps(), get_eel(), get_dsf()

    calc. parameters: param[<key>]
      <key>:  'filename', 'quantity', 'lfe', 'kernel',
              'npwwfn', 'nbands', 'npwmat', 'emin', 'emax', 'broad',
              'lomo', 'nkbz', 'nkibz', 'dsf_fact'
              'iqpol', 'q_rc', 'q_cc', 'qred_rc', 'qred_cc', 'G_rc', 'G_cc' 
   
"""
import sys; 
import numpy as np;
from   scipy.linalg import norm;
# http://sourceforge.net/projects/graceplot/
import GracePlot as gp;
import dp_mdf;
from dp2plot import *;

      
#=== 1. Selection of Spectra ===========================================
#
#--- 1.1. Read DP files (list of MDF-objects) --------------------------

spectra=GetMDF("tests/9pol*.mdf");
#spectra.extend(GetMDF("bla"));
y="eels";   # Quantity for y-axis: "eels", "ieps", "reps"

#--- 1.2. Add parameters (Optional) ------------------------------------

for mdf in spectra:
  qr=mdf.param['qred_cc']; G=mdf.param['G_cc'];
  #mdf.param['alpha'] = 180/3.14159*np.arccos(np.dot(G,qr)/norm(G)/norm(qr));

#--- 1.3. Filter Spectra (Optional) ------------------------------------

def select_spectrum(mdf):
  "RETRUN 'False' if spectrum should be filtered out"
  ret=True;
  #if ( max(abs(mdf.get_q()-np.array([2,2,2])))  > 0.1 ):  ret=False;
  if ( mdf.param['lfe']   <> 'lf' ):  ret=False;
  return ret;

spectra=filter(select_spectrum,spectra);

#--- 1.4. Sort Spectra (Optional) --------------------------------------

spectra.sort(key=lambda mdf: norm(mdf.param['npwmat']));
PrintFilenames(spectra);




#=== 2. How to plot each Spectrum ======================================
#
#--- 2.1. XY-Data ------------------------------------------------------

def get_x(mdf):
  return mdf.get_E();

def get_y(mdf):
  if (y=="dsf"):  return mdf.get_dsf()*1000;
  if (y=="eels"): return mdf.get_eel();
  if (y=="reps"): return mdf.get_Reps();
  if (y=="ieps"): return mdf.get_Ieps();

#--- 2.2. Legend -------------------------------------------------------

def get_legend(mdf):
  return "qpol=%d"%(mdf.param['iqpol']);

#--- 2.3. Colors -------------------------------------------------------

NUM_COLORS = 6;

def get_linecolor(mdf):
  "RETURN the number of the color for spectrum mdf"
  get_linecolor.counter+=1; # number of calls of this function
  return 2+get_linecolor.counter%6;
get_linecolor.counter=-1;

def SetPalette(fig):
  "re-define colors except white (0) and black (1)"
  i=2;
  for x in map(int, np.linspace(0,255,NUM_COLORS)):
    fig.assign_color(i, (x, 0, 255-x), "color%d"%(i));
    i+=1;

#--- 2.4. Line Properties ----------------------------------------------

def get_linestyle(mdf):
  if mdf.param['iqpol']%3 == 2: return 2;
  return 1;




#=== 3. Graph Properties ===============================================
#
#--- 3.1. Creation of Graph --------------------------------------------

fig = gp.GracePlot(width=8, height=6 );
SetPalette(fig);

ax  = fig[0];
ax.SetView(xmax=1.2,ymax=0.85,aspect_scaled=False);
ax.plot([gp.Data(x=get_x(mdf), y=get_y(mdf),
                 legend=get_legend(mdf),
                 line=gp.Line(linestyle=get_linestyle(mdf),color=get_linecolor(mdf)))
         for mdf in spectra]);
ax.legend();

#--- 3.2. Graph Properties ----------------------------------------------

mdf = spectra[-1]; # use last spectrum for setting parameters
ax.title("Graphite, RPA, q=(%d,%d,%d)\src\N"
         % tuple(mdf.param['G_rc']),font=4,size=2);
  
ax.subtitle("(job:%s, pm%d, bd%d, pw%d, lo%d, broad %3.2f )"
         % (mdf.param['filename'], mdf.param['npwmat'], mdf.param['nbands'],
            mdf.param['npwwfn'],   mdf.param['lomo'],   mdf.param['broad']) );

#--- 3.3. Axis Properties -----------------------------------------------

ax.xaxis(xmax=22.0001,label=gp.Label('Energy [eV]',font=4,charsize=1.5))
if y == "reps": ax.yaxis(label=gp.Label('Re \\f{Symbol} e \\f{} [a.u.]',font=4,charsize=1.5))
if y == "ieps": ax.yaxis(label=gp.Label('Im \\f{Symbol} e \\f{} [a.u.]',font=4,charsize=1.5))
if y == "eels": ax.yaxis(label=gp.Label('- Im \\f{Symbol} 1/e \\f{} [a.u.]',font=4,charsize=1.5))
if y == "dsf" : ax.yaxis(label=gp.Label('S(\\f{Symbol}w\\f{},q)   [1/keV]',font=4,charsize=1.5))  
  
fig.redraw(force=True)


#=== 4. Save file ========================================================

#outfile="Test-q%4.3f-%4.3f-%4.3f" % tuple(mdf.param["q_rc"]);
#print "Save to name ",outfile;
#fig.save(outfile, format="postscript");
#fig.save(outfile,format="agr");

## NOTE: add the following code to the GracePlot.save() routine in GracePlot.py
## elif ext == '.ps': 
##   self.send_commands('PAGE RESIZE 840,594','hardcopy device "PostScript"',
##                      'print to "%s"' % filename, 'print')

#=== 5. Debugging ========================================================

# PyPlot(spectra);

