"""
     Input/Output module for binary chi0-files generated by DP 
     using the 'writechi0' flag (available with dpforexc, r1569)
     
   VERSION

     $Id: dp_chi0.py 366 2011-11-10 11:34:46Z hambach $

   USAGE

     # read output file (header + data)
     data = DPchi0("tests/1pol_chi0.dat",verbosity=3);

     # get chi0 on a hypercube (for energy 2.4eV)
     io=data.E.index([2.4])[0];
     (chi0,G_grid)=data.get_chi0_as_hypercube(io,Gmax=(None,None,1));

   TODO
     - iterate over energies without reading whole file at once
     - speed up reindexation in get_chi0_as_hypercube (avoid explicit loop)
     - speed up translate ()

   COPYRIGHT

     Copyright (c) 2011, Ralf Hambach. All rights reserved.
     Use of this source code is governed by a BSD-style license that 
     can be found in the LICENSE file.
"""
__version__ = "$Revision: 343 $"
__author__ = "Ralf Hambach"
__date__ = "$Date: 2011-05-19 01:06:46 +0200 (Thu, 19 May 2011) $"

import numpy as np;
import numpy.linalg;
import sys;
import fortranfile as ff;
import SRcommon
import dp_mdf

def verify_equal(a,b,delta=1e-6):
  diff = np.abs(a-b); 
  if np.max(diff)>delta:
    raise ValueError('ERROR! found absolute difference ' +\
      str(np.max(diff))+' > '+ str(delta)+ \
      '\n expected equal values for a='+str(a)+'\n and b='+str(b));

def verify_equal_rel(a,b,delta_rel=1e-6):
  diff = np.abs(a-b); av = (np.abs(a)+np.abs(b))/2.;
  if np.max(diff/av)>delta_rel: 
    raise ValueError('ERROR! found relative difference '+\
      str(np.max(diff/av))+' > '+ str(delta_rel)+ \
      '\n expected equal values for a='+str(a)+'\n and b='+str(b));

class DPchi0:
  """
  reads the chi0 file written by DP with the 'chi0->py' flag
  and provides different views of the data, in particular

    -> all calculation parameters given in the header
    -> the raw-data for all polarization directions

  Internal data
    
    E    ... grid of energies (RegularVGrid)
    chi0 ... complex array (nomega,ngvec,ngvec)
    param... dictionary with calculation parameters from DP, keys:

      nomega, emin, emax, estep, npwmat, npwwfn, nsppol,
      nbands, lomo, lumo, broad, tpula,kBT, DPflags(21),
      soenergy, stretchvalence, stretchconduction, nel
      nkibz, nkbz, bzvol, ucvol, rprimd(3,3), gprimd(3,3), 
      ngvec, gvec(ngvec,3), ig, G_rc(3), G_cc(3), nqpol, 
      q_rc, qred_rc, q_cc, qred_cc [of shape (3,) or (nqpol,3)]
      nsym, nop, nopq, nopr, ninv, ninvq, 
      sym_r(nsym,3,3), sym_t(nsym,3), sym_tr(nsym),  
      natoms, T(natoms), 
 
  only if nqpol>1: 
    uwing,lwing... complex array (nomega,nqpol,ngvec)
    head       ... complex array (nomega,3,3)
  """

  def __init__(self,filename,npwmat=None,verbosity=1,**kwargs):
    """
    filename ... full name of the inputfile
    npwmat   ... (opt) set default for npwmat (can be overwritten later on)
    verbosity... (opt) 0 silent, 1 few output, 2 more output, 3 debugging
    
    further arguments:
    endian   ... define the endianness of the input file ">", "<"
    """
    
    self.filename = filename;
    self.verbosity= verbosity;

    if verbosity>0: print "\n--- read chi0 -----------------------------------------";

    try:
      file = ff.FortranFile(filename,**kwargs);
      self.__scan_header(file);
      self.__scan_data(file);     
    except:
      print "\nERROR! while reading file '%s'\n\n"%(filename)
      raise

    if npwmat is None: npwmat = self.param['npwmat'];
    self.npwmat = npwmat

    # DEBUG: cross check with parameters from outnlf-file
    if verbosity>2: 
      outnlf = filename.replace("chi0.dat","outnlf.mdf");
      self.__check_outnlf(outnlf);

    if verbosity>1: print "-------------------------------------------------------"


  def __scan_header(self,f):

    HARTREE=27.2113845 # standard uncertainty 0.0000023 (NIST)

    # comment
    try:
      comment=f.readString();
    except IOError:
      raise IOError(" Cannot read input file.\n          "+\
                    "Try to change parameter 'endian' in creation of DPchi0.");
    self.__header_version, = f.readInts();
    if self.verbosity > 2: 
      print "\n "+comment;
      print " header version ",self.__header_version;
    
    # calculation flags
    p=dict(); d=dict();
    d['TESTPARTICLE'], d['TESTELECTRON'], d['JDOS'], d['RPA'], d['LDA'], d['GW'],   \
    d['SO'], d['NLDA'], d['qps'],d['qpsenergy'], d['KP_THEORY'], d['kernel'],       \
    d['lifetime'], d['DRUDE'], d['ONLYDRUDE'], d['inv'], d['td'], d['xx'], d['vkb'],\
    d['hilbert'], d['metal'] = map(bool,f.readInts());
    p['DPflags']=d;

    # basic calculation parameters
    p['npwmat'], p['npwwfn'], p['nsppol'], p['nomega'], p['nbands'], \
    p['lomo'], p['lumo'], p['nkibz'], p['nkbz'], p['nel'], p['nop'], p['nopr'],  \
    p['nopq'], p['ninv'], p['ninvq'] = f.readInts();
    p['soenergy'],p['stretchvalence'],p['stretchconduction'],p['kBT'] = f.readReals();

    # energies (in eV)
    p['emin'],p['emax'],p['estep'],p['broad'] = f.readReals()*HARTREE;
    emin=p['emin']; emax=p['emin']+p['estep']*(p['nomega']-1);
    self.E = SRcommon.RegularVGrid([p['nomega']],[[emin,emax]],('Energy',),('eV',));

    # q/G-vectors (in [r.c.])
    p['nqpol'],p['ig']=f.readInts();
    p['G_rc']  = f.readInts();
    shape = (p['nqpol'],3) if p['nqpol'] > 1 else (3,);
    p['qred_rc']=f.readReals().reshape(shape);
    p['gvec']  = f.readInts().reshape(p['npwmat'],3);
    p['ngvec'] = p['npwmat'];

    # crystal structure
    #                    rprimd = [[a1],    gprimd = [[b1],
    #                    in [au]   [a2],    in [au]   [b2],      
    #                              [a3]]              [b3]]   
    p['ucvol'],p['bzvol'],p['tpula']=f.readReals();
    p['rprimd']= f.readReals().reshape(3,3);
    p['gprimd']= f.readReals().reshape(3,3);
    gprimd = 2*np.pi*numpy.linalg.inv(p['rprimd']).T; # a_i.b_j=2pi*delta_ij
    assert(np.all(np.abs(gprimd-p['gprimd'])<1e-6));
   
    # chemistry (in [a.u.])
    p['natoms'],= f.readInts();
    znucl = f.readReals(prec='d');
    xcart = f.readReals(prec='d').reshape(p['natoms'],3);
    p['T'] = []; 
    for i in range(p['natoms']):
      typeat=np.where(np.unique(znucl)==znucl[i])[0];
      p['T'].append(SRcommon.Atom(znucl[i],xcart[i],typeat));
      
    # symmetry (rotation, translation [r.c.], time-reversal)
    p['nsym'], = f.readInts();
    p['sym_r'] = f.readInts().reshape(p['nsym'],3,3);
    p['sym_t'] = f.readReals(prec='d').reshape(p['nsym'],3);
    p['sym_tr']= f.readInts();
    
    # end of header
    if f.readInts()[0]<>11111:
      raise IOError("could not read full header");

    # calculate missing parameters
    p['qred_cc']=np.dot(p['qred_rc'],p['gprimd']);
    p['G_cc']   =np.dot(p['G_rc']   ,p['gprimd']);
    p['q_rc']   =p['qred_rc']+p['G_rc'];
    p['q_cc']   =p['qred_cc']+p['G_cc'];
    p['dsf_fact']= np.sum(p['q_cc']**2)/(4 * np.pi**2 * p['nel']/p['ucvol']);
    self.param=p;

    # determine if wings are included in chi0-file
    self.has_wings =     self.param['DPflags']['KP_THEORY'] and \
                     not self.param['DPflags']['ONLYDRUDE'];

    if self.verbosity>1:
      self.write_header();


  def __scan_data(self,f):
    """
    read chi0 at once from the input file:
       chi0(nomega,ngvec,ngvec)
       lwing(nomega,nqpol,ngvec)  (if present)
       uwing(nomega,nqpol,ngvec)  (if present)
       head(nomega,3,3)           (if present)
    TODO: implement spin-polarised case
    """
    # abbreviations
    nomega = self.E.size; ngvec=self.param['ngvec']; nqpol=self.param['nqpol'];
    if self.param['nsppol']>1: 
      raise NotImplemented("Spin-dependent chi0 not jet implemented");

    # allocate memory
    chi0=np.empty((nomega,ngvec,ngvec),dtype=complex);
    if self.has_wings:
      lwing=np.empty((nomega,nqpol,ngvec),dtype=complex);
      uwing=np.empty((nomega,nqpol,ngvec),dtype=complex);
      head =np.empty((nomega,3,3),dtype=complex);

    # loop over energies
    for io in range(nomega):

       # body: chi0(io,ig,igp)
       data=f.readReals().reshape(ngvec,ngvec,2)
       chi0[io]=data[:,:,0]+data[:,:,1]*1.j;

       if self.has_wings:
         # wings: wing(io,iq,ig)
         data=f.readReals().reshape(nqpol,ngvec,2);
         lwing[io]=data[:,:,0]+data[:,:,1]*1.j;
         data=f.readReals().reshape(nqpol,ngvec,2);
         uwing[io]=data[:,:,0]+data[:,:,1]*1.j;
         
         # head: head(io,ip,ipp)
         data=f.readReals().reshape(3,3,2);
         head[io]=data[:,:,0]+data[:,:,1]*1.j;

    # end of body
    if f.readInts()[0]<>22222:
      raise IOError("could not read full body");

    # copy wings to chi0 if nqpol=1 (k.p with one selected q)
    if self.has_wings:
      if nqpol==1:
        chi0[:,:,0] = lwing[:,0,:];
        chi0[:,0,:] = uwing[:,0,:];
      else:
        self.lwing= lwing;
        self.uwing= uwing;
        self.head = head;

    # return
    self.chi0 = chi0;



  def __check_outnlf(self,filename):
    """ 
    cross-check with header from corresponding outnlf-file
    """
    DPmdf = dp_mdf.DPmdf(filename);
    p     = DPmdf.param;

    print " Cross-checking parameters from 'chi0.dat' and 'outnlf.mdf' ...";
 
    # copy q-vectors to param (flatten array if nqpol==1)
    p['nqpol'] = nqpol = DPmdf.nqpol;
    shape = (nqpol,3) if nqpol > 1 else (3,);
    for key in ('q_rc','q_cc','qred_rc','qred_cc'):
      p[key] = np.array(DPmdf.__dict__[key],float).reshape(shape);
    for key in ('G_rc','G_cc'):
      p[key] = np.array(DPmdf.__dict__[key],float);
    #print 'Parameters from outnlf.mdf:\n', ", ".join(np.sort(p.keys()));
    #print 'Parameters from chi0.dat:\n', ", ".join(np.sort(self.param.keys()));

    # compare parameters
    common_keys=set(p.keys()).intersection(self.param.keys());
    print "\n parameter  max-diff"
    for key in common_keys:
      diff = np.max(np.abs(p[key]-self.param[key]))
      print "  %s  %.1g"%(key.ljust(10),diff);
      if diff>1e-6:
        raise IOError(" Parameter %s differs! \n"%key+\
                      str(p[key])+"\n\n"+str(self.param[key]));
    print "\n NOT CHECKED\n  "+\
      ", ".join(np.sort(list(set(self.param.keys()).difference(common_keys))));

    # compare data
    print "\n compare energies ...",;
    verify_equal(self.E.flatten(),DPmdf.E,delta=1e-6);
    print " OK";

    print " compare data ... pol ",;
    for iq in range(p['nqpol']):
      print "%d, "%iq,
      q_cc=self.param['q_cc'].reshape(p['nqpol'],3)[iq];
      ig  =self.param['ig']-1; # fortran index
      eps1=np.asarray(DPmdf.Reps[iq]) + 1.j*np.asarray(DPmdf.Ieps[iq]);
      chi0=[self.get_chi0(io,iqpol=iq)[0][ig,ig]  \
               for io in range(self.param['nomega'])];
      eps2=1.-4*np.pi/(q_cc**2).sum() * np.asarray(chi0);
      verify_equal_rel(eps2, eps1,delta_rel=1e-3);
    print " OK\n"

  def write_header(self):
    " write parameters from the header of the chi0.dat file "

    p=self.param;
    print "\n Crystal structure ";
    print "   lattice vectors [a.u.]:            "; print(p['rprimd']);
    print "   reciprocal lattice vectors [a.u.]: "; print(p['gprimd']);
    for i in range(p['natoms']):    
      print "   atom of %s"%p['T'][-1];
    print " ";
    print " DP-parameters ";
    print "   energy: %8f - %8f eV, dE=%8f eV "%(p['emin'], p['emax'], p['estep']);
    print "   q [rc]: \n",p['q_rc'].round(6);
    print "   q [cc]: \n",p['q_cc'].round(6);
    print "   ngvec : ",p['ngvec'];
    print "   nqpol : ",p['nqpol'];
    print " ";

  def get_shells(self,gvec=None,gprimd=None):
    """
    return list of indices for closed shells in G-vectors
    """
    if gvec is None:   gvec  =self.param['gvec']; 
    if gprimd is None: gprimd=self.param['gprimd'];
    
    glen  =np.sqrt(np.sum(np.dot(gvec, gprimd)**2,axis=1));
    shells=np.where(glen[1:]-glen[:-1]>1e-6)[0]+1;

    return np.append(shells,len(gvec));

  def print_shells(self,out=sys.stdout,gvec=None,gprimd=None):
    """
    print list of closed shells to file out
    """
    if gvec is None:   gvec  =self.param['gvec']; 
    if gprimd is None: gprimd=self.param['gprimd'];
 
    shells=self.get_shells(gvec,gprimd);
    out.write(" npwmat  |G|   G-vectors [r.c.]\n");
    for i in range(len(shells)-1):
      glen = np.sqrt(np.sum(np.dot(gvec[shells[i]], gprimd)**2));
      out.write("\n  %3d  %5.2f "%(shells[i+1],glen));
      for (n,ig) in enumerate(range(shells[i],shells[i+1])):
        if (n%5==0 and n>0): out.write("\n             ");
        out.write(" [%2d,%2d,%2d]"%tuple(gvec[ig]));
    out.write("\n");
     

  def get_chi0(self,io,iqpol=0,npwmat=None):
    """
    returns chi0 for a given frequency and polarisation direction
    as 2D-array chi0(ig,ig) like in DP

    INPUT
     io   ... Energy index, e.g., io=E.index([2.4])[0]; # for 2.4 eV
     iqpol... (opt) polarisation direction (iqpol<nqpol)
     npwmat.. (opt) restrict chi0 to the first npwmat G-vectors

    OUTPUT
     chi0 ... 2D-array chi0(ngvec,ngvec)
     gvec ... reference to array of G-vectors gvec(ngvec,3)
    """
    gvec  =self.param['gvec']; 
    if io    >= self.param['nomega']:raise ValueError('"io" out of range'); 
    if iqpol >= self.param['nqpol']: raise ValueError('"iqpol" out of range');
    if npwmat is None:               npwmat=self.npwmat;
    if npwmat not in self.get_shells():
      self.print_shells();
      raise ValueError("ERROR! shell not closed at npwmat=%d (see list above)."%npwmat);

    chi0 = self.chi0[io,:,:];
    if self.param['DPflags']['KP_THEORY'] and self.param['nqpol']>1:
      chi0[:,0] = self.lwing[io,iqpol,:];
      chi0[0,:] = self.uwing[io,iqpol,:];
    return (chi0[:npwmat,:npwmat], gvec[:npwmat]);
  
  def get_chi0_as_hypercube(self,io,iqpol=0,Gmax=(None,None,None),npwmat=None,verbosity=None):
    """
    returns chi0 for a given frequency and polarisation direction
    as 6D-array indexed by the three components of the G-vectors
  
     chi0_cube(iG1,iG2,iG3; iG1',iG2',iG3')
  
    The origin is in the center of the cube. By default, the size of the
    cube is chosen such that it contains all G-vectors in gvec (they
    form a sphere in cartesian coord.). Unknown elements are set to zero.
    If one of the Gmax components is given, the cube contains G-vectors 
    with components -Gmax[i],..0,..Gmax[i].
  
    INPUT
     io   ... Energy index, e.g., io=E.index([2.4])[0]; # for 2.4 eV
     iqpol... (opt) polarisation direction (iqpol<nqpol)
     Gmax ... (opt) [Gmax1,Gmax2,Gmax3] specifies the 
                    size of the cube for each direction in r.c.
     npwmat.. (opt) restrict chi0 to the first npwmat G-vectors
     verbosity...(opt) defines debugging output
  
    OUTPUT
     chi0_cube ... 6D-array chi0_cube(Gmax[0],Gmax[1],Gmax[2]; Gmax[0],Gmax[1],Gmax[2])
     G_grid    ... corresponding G-vectors as RegularVGrid
    """
    if verbosity is None: verbosity=self.verbosity;  
    (chi0,gvec) = self.get_chi0(io,iqpol,npwmat=npwmat);

    # define cube dimensions
    Gmax_in  =np.asarray(Gmax);
    Gmax_chi0=np.abs(gvec[:]).max(axis=0).round();
    Gmax     =np.where(Gmax_in >= 0,Gmax_in,Gmax_chi0);

    # reduce chi0 according to cube dimensions
    if any(Gmax<Gmax_chi0):
      ig = np.where( np.all( np.abs(gvec)<=Gmax, axis=1 ))[0];
      chi0 = chi0[np.ix_(ig,ig)];
      gvec = gvec[ig];
      if verbosity>0:
        # calculate closed shells with G<=Gmax 
        shells    = self.get_shells();
        ig_closed = ig[ ig[1:]-ig[:-1]>1 ][0] + 1;
        sh_closed = shells[shells <= ig_closed][-1];
        assert( np.all(gvec[:sh_closed]==self.param['gvec'][:sh_closed]) );
        print('get_chi0_as_hypercube(): restricting G-vectors ' + \
              'to |G| < (%d,%d,%d).\n'% tuple(Gmax)             + \
              'WARNING: This will break closed shells! Use npwmat=%d instead.'%(sh_closed));
    ngvec= gvec.shape[0];
  
    # rectangular, symmetric cube around G-vectors
    Gmax  = np.asarray(Gmax,dtype=int);
    extent= np.array([-Gmax,Gmax]);
    num   = 2*Gmax+1;
    G_grid =SRcommon.RegularVGrid(num,extent.T,('G1','G2','G3'),('G [r.c.]',));
    G_index=np.array([ G_grid.index(gvec[ig]) for ig in range(ngvec) ]);
  
    # re-index chi0_AO(ig,ig') as chi0_AO(G1,G2,G3,G1',G2',G3');
    chi0_cube=np.zeros((G_grid.shape+G_grid.shape),dtype=complex);
    for ig in range(ngvec):
      for igp in range(ngvec):
        # build index for chi0: might be faster
        index=G_index[ig].tolist()+G_index[igp].tolist();
        chi0_cube[tuple(index)] = chi0[ig,igp]; # copy ?

    if verbosity>2: 
      print('get_chi0_as_hypercube(): Gmax = (%3d,%3d,%3d)'%tuple(Gmax));

  
    return(chi0_cube,G_grid);


  def plot_GzGz(self,(Gx,Gy),(Gxp,Gyp)):
    import SRplot;

    chi0_GzGz=[];
    for io in range(self.param['nomega']):

      (chi0,G_grid) = self.get_chi0_as_hypercube(io,verbosity=0);
      iG = G_grid.index([Gx,Gy,0]);
      iGp= G_grid.index([Gxp,Gyp,0]);
      chi0_GzGz.append(chi0[iG[0],iG[1],:,iGp[0],iGp[1],:]);

    Gz=G_grid.max[2]+1e-6; L=G_grid.shape[2];
    BULK_G_slice=SRcommon.RegularVGrid([L,L,1],[[-Gz,Gz],[-Gz,Gz],[0,0]],('Gz','Gz''','none'),('G [r.c.]'));
    SRplot.imshow3D_python(np.asarray(chi0_GzGz).imag,self.E,BULK_G_slice);


  def plot_zz(self,(Gx,Gy),(Gxp,Gyp)):
    import SRplot;
    import numpy.fft as fft;

    chi0_zz=[];
    for io in range(self.param['nomega']):

      (chi0,G_grid) = self.get_chi0_as_hypercube(io,verbosity=0);
      iG = G_grid.index([Gx,Gy,0]);
      iGp= G_grid.index([Gxp,Gyp,0]);
      L  = G_grid.shape[2];
      chi0_zz.append( fft.fftshift(fft.ifft2(fft.ifftshift( \
         chi0[iG[0],iG[1],:,iGp[0],iGp[1],::-1])))*L );

    zmax=self.param['rprimd'][2,2]/2;
    BULK_z_slice=SRcommon.RegularVGrid([L,L,1],[[-zmax,zmax],[-zmax,zmax],[0,0]],('z [a.u.]','z'' [a.u.]','none'),('G [r.c.]'));
    SRplot.imshow3D_python(np.asarray(chi0_zz).imag,self.E,BULK_z_slice);


  def translate(self,tau):
    """
    Change phase of chi0 according to a shift of the system by the 
    vector tau in real-space: phase = exp( -2pi j (G-G').tau )

     tau ... real-space vector in [r.c.]

    """
    gvec=self.param['gvec'];
    for ig in range(len(gvec)):
      for igp in range(len(gvec)):
         phase = np.dot(gvec[ig]-gvec[igp], tau);
         self.chi0[:,ig,igp] *=  np.exp( -2.j*np.pi*phase );

    if self.param['nqpol']>1:
      for ig in range(len(gvec)):
        phase = np.dot(gvec[ig], tau);
        self.lwing[:,:,ig] *= np.exp( -2.j*np.pi*phase );
        self.uwing[:,:,ig] *= np.exp(  2.j*np.pi*phase ); # igp !

  def inversion(self):
    """
    Real space inversion of chi0 [assuming that chi0(q+G,q+G') = chi0(-q+G,-q+G')]  
    """
    raise NotImplemented();
    #index= [ np.where( np.all(self.ngvec == -self.ngvec[ig]) for ig in range(self.ngvec)];
    self.chi0 = self.chi0[index,index];
    
    if self.nqpol>1:  
      self.lwing = self.lwing[:,:,index];
      self.uwing = self.uwing[:,:,index];

  def sum(self, other, scale=1.):
    # unit-cell and reciprocal space must be the same
    verify_equal(self.param['rprimd'],other.param['rprimd']);
    verify_equal(self.param['gprimd'],other.param['gprimd']);
    verify_equal(self.param['gvec']  ,other.param['gvec']  );
    
    self.chi0 += other.chi0; self.chi0 *= scale;
    if self.param['nqpol']>1:
      self.lwing += other.lwing; self.lwing *= scale;
      self.uwing += other.uwing; self.uwing *= scale;


### Self-Tests
if __name__ == '__main__':

   # 6-polarisation (k.p calculation)
   # read chi0 file (big endian)
   data = DPchi0("./tests/chi0/9pol_chi0.dat",endian='>',verbosity=3);

   # 1-polarisation (q>0)
   # read chi0 file (big endian)
   DPchi0 = DPchi0("./tests/chi0/1pol_chi0.dat",endian='>',verbosity=3);

   # test write outnlf.mdf file
   E=DPchi0.E.flatten();
   q_cc=DPchi0.param['q_cc']; ig=DPchi0.param['ig']-1; 
   eps =1.-4*np.pi/(q_cc**2).sum() * DPchi0.chi0[:,ig,ig];
   mdf=dp_mdf.MDF(DPchi0.param,{'lfe': 'nlf', 'kernel': 'rpa'},E,eps);
   out = open('tmp.mdf','w'); mdf.write(out); out.close();

   # test get_chi0_as_hypercube (for first energy)
   io=DPchi0.E.index([2.4])[0];
   (chi0,G_grid) = DPchi0.get_chi0_as_hypercube(io,Gmax=(None,2,4),npwmat=25);

   # test plot
   import matplotlib.pylab as plt;
   DPchi0.plot_GzGz((0,0),(0,0));
   DPchi0.plot_zz((0,0),(0,0));

   plt.show();

