'''
Extract following info from DP-logfile:
 - KSS (eigenvalues, k-points)
 - occupation numbers
 - q and G-vectors

Needs verboselevel > 1
'''

import numpy as np
import re;
import string;
import sys

def search_line(file,pattern):
  while True:
    try:
      line=file.next();
      #print line;
    except StopIteration: 
      #print "EOF reached";
      return None;
  
    res=re.findall(pattern,line)
    if (res):
      print "PARSED: "+line,
      return res;

def read_wrapped_line(file,num):
  """ read num values ignoring line breaks"""
  data = file.next().split();
  while (len(data)<num):
    data.extend(file.next().split());
  if (len(data)>num):
    print "FORMAT ERROR! read_wrapped_line() got more values than expected.";
    print data;
  return data;


class DPlog:
  "Parse log information of DP"

  def __init__(self,file):
    "file ... file handle of the logfile"
    self.ReadKSS(file);
    self.ReadQandG(file);
    self.ReadFermi(file);

  def ReadKSS(self,file):

    # read kss filename (might be line wrapped)
    line=search_line(file, "^ \*{9} Checking KS structure file .*")[0];
    res=re.findall("^ \*{9} Checking KS structure file (\S+)",line);
    if(not res):
      self.filename= read_wrapped_line(file,1)[0]
    else:
      self.filename= res[0]
    
    # get number of bands
    self.nbands=int(search_line(file,\
        "^ number of bands included, \s* nbands = \s* ([\d]+)")[0]);
    # read kpoints in r.c.
    self.nkibz=int(search_line(file, \
        "^  number of k-points \s* ([\d]+)")[0]);
    file.next();
    self.kibz_rc = np.empty((self.nkibz,3),dtype=float);
    for ik in range(self.nkibz):
      line=file.next();
      self.kibz_rc[ik]=map(float,line.split());

    # read eigenvalues in eV
    search_line(file,"^  k       eigenvalues \[eV\]");
    self.E_kn=np.empty((self.nkibz,self.nbands),dtype=float);
    for ik in range(self.nkibz):
      data=read_wrapped_line(file,self.nbands+1);
      self.E_kn[ik] = map(float,data)[1:];

    # end of kss
    search_line(file,"^ \*{40}");


  def ReadQandG(self,file):

    # read block of information:
    search_line(file, "^ \*+ about q and G \*+");
    lines=[file.next()];
    while not re.match("^ q-points check ok",lines[-1]):
      lines.append(file.next());
    lfile=iter(lines);

    # real and reciprocal lattice vectors
    if search_line(lfile, "^ real-space lattice vectors "):
      rprimd=read_wrapped_line(lfile,9);
      self.rprimd=np.asarray(rprimd,dtype=float).reshape(3,3);

    if search_line(lfile, "^ calculated reciprocal lattice vectors "):
      gprimd=read_wrapped_line(lfile,9);
      self.gprimd=np.asarray(gprimd,dtype=float).reshape(3,3);
      #print self.rprimd,self.gprimd;
    lfile=iter(lines); # start over again

    # Get q: "# q = ( 0.200000, 0.000000,-0.113636) [r.l.u.]"
    #        "# q = ( 0.068728,-0.039680,-0.154474) [c.c.]  "
    #        "# q = (  .176777,  .102062,  .000000) [c.c. units of 2pi/a =   .959379 a.u.]"
    def rd_vec(name,unit):
      return search_line(lfile,"^ "+name+"\s*=\s*"+\
        "\(\s*(-*[\d\.]+),\s*(-*[\d\.]+),\s*(-*[\d\.]+)\s*\)\s*"+unit);

    self.nqpol=0; self.qred_rc=[];self.qred_cc=[];
    while True:
      qred_rc = rd_vec("q","\[r\.l\.u\.\]");
      qred_cc = rd_vec("q","\[c\.c\.\]");
      qred_cc2= rd_vec("q","\[c\.c\. units of 2pi/a .*\]");
      if qred_rc is None: break;
       
      self.nqpol+=1;
      self.qred_rc.append(map(float,qred_rc[0]));
      self.qred_cc.append(map(float,qred_cc[0]));

    # convert to array
    self.qred_rc = np.asarray(self.qred_rc,float);
    self.qred_cc = np.asarray(self.qred_cc,float);
    lfile=iter(lines); # start over again


    # Get G: "# G = (  0,  0,  0) [r.l.u.]"
    #        "# G = (  0.000000,  0.000000,  0.000000) [c.c.]"
    #        "# G = (  .000000,  .000000,  .520275) [c.c. units of 2pi/a =   .959379 a.u.]"
    search_line(lfile, "^ calculating at the following G:");
    self.G_rc=np.asarray(rd_vec("G","\[r\.l\.u\.\]")[0],dtype=float);
    self.G_cc=np.asarray(rd_vec("G","\[c\.c\.\]"   )[0],dtype=float);

    # calculate total q:
    self.q_rc = self.G_rc+self.qred_rc;
    self.q_cc = self.G_cc+self.qred_cc;

    #print self.q_rc, self.qred_cc, self.G_rc;

  def ReadFermi(self,file):
    search_line(file, "^ \*+ Fermi level analysis \*+");

    # Fermi energy
    (Ef,)=search_line(file, "^ Fermi energy = \s*(-*[\d\.]+) \[eV\]");
    self.Efermi=float(Ef);

    # read new occupation numbers
    search_line(file, "^ Occupation numbers after recalculation:");
    self.f_kn=np.empty((self.nkibz,self.nbands),dtype=float);
    for ik in range(self.nkibz):
      data=read_wrapped_line(file,self.nbands+1);
      self.f_kn[ik] = map(float,data)[1:];


if __name__ == "__main__":
  import matplotlib.pylab as plt;
  import mpl_toolkits.mplot3d as p3d;

  # READ FILE
  #file=open("dp.out");
  file=open("./tests/hybrid_K_kp_fast_out");
  log=DPlog(file);
  file.close();
  
  gap=log.E_kn[:,8]-log.E_kn[:,7];
  
  # show gap for a cut in xy-plane of BZ
  (iXY,)=np.where(log.kibz_rc[:,2]==0.0625)

  fig=plt.figure();
  plt.scatter(log.kibz_rc[iXY][:,0],log.kibz_rc[iXY][:,1],c=gap[iXY]);
  #ax =p3d.Axes3D(fig);
  #ax.scatter3D(log.kibz_rc[iXY][:,0],log.kibz_rc[iXY][:,1],gap[iXY]);
  plt.show();

  (iXY,)=np.where(gap<2)
  # output for gnuplot
  file=open("gnuplot.dat",'w');
  for ik in range(len(iXY)):
    file.write(" %f %f %f \n"%tuple(log.kibz_rc[iXY[ik],:]));
  file.close();

